@@ -8,16 +8,21 @@
 #include <libgnomeui/gnome-thumbnail.h>
 #include <gdk-pixbuf/gdk-pixbuf.h>
 #include <libgnomevfs/gnome-vfs.h>
+#include <eel/eel-vfs-extensions.h>
 #if HAVE_EXIF
 #include <libexif/exif-data.h>
 #include <libexif/exif-utils.h>
+#include <libexif/exif-loader.h>
 #endif
 
 #include "libeog-marshal.h"
 #include "eog-image.h"
 #include "eog-image-private.h"
 #include "eog-pixbuf-util.h"
+#include "eog-image-cache.h"
+#if HAVE_JPEG
 #include "eog-image-jpeg.h"
+#endif
 
 static GThread     *thread                     = NULL;
 static gboolean     thread_running             = FALSE;
@@ -289,16 +294,13 @@ eog_image_dispose (GObject *object)
 
 	priv = EOG_IMAGE (object)->priv;
 
+	eog_image_free_mem (EOG_IMAGE (object));
+
 	if (priv->uri) {
 		gnome_vfs_uri_unref (priv->uri);
 		priv->uri = NULL;
 	}
 
-	if (priv->image) {
-		g_object_unref (priv->image);
-		priv->image = NULL;
-	}
-
 	if (priv->caption) {
 		g_free (priv->caption);
 		priv->caption = NULL;
@@ -309,6 +311,11 @@ eog_image_dispose (GObject *object)
 		priv->caption_key = NULL;
 	}
 
+	if (priv->file_type) {
+		g_free (priv->file_type);
+		priv->file_type = NULL;
+	}
+	
 	if (priv->status_mutex) {
 		g_mutex_free (priv->status_mutex);
 		priv->status_mutex = NULL;
@@ -327,13 +334,6 @@ eog_image_dispose (GObject *object)
 		g_list_free (priv->undo_stack);
 		priv->undo_stack = NULL;
 	}
-
-#if HAVE_EXIF
-	if (priv->exif) {
-		exif_data_unref (priv->exif);
-		priv->exif = NULL;
-	}
-#endif
 }
 
 static void
@@ -493,6 +493,7 @@ eog_image_new_uri (GnomeVFSURI *uri)
 
 	priv->uri = gnome_vfs_uri_ref (uri);
 	priv->mode = EOG_IMAGE_LOAD_DEFAULT;
+	priv->status = EOG_IMAGE_STATUS_UNKNOWN;
 	priv->modified = FALSE;
 	priv->load_thread = NULL;
 
@@ -716,124 +717,6 @@ update_exif_data (EogImage *image)
 	}	
 }
 
-#define JPEG_MARKER_SOI  0xd8
-#define	JPEG_MARKER_APP0 0xe0
-#define JPEG_MARKER_APP1 0xe1
-
-typedef enum {
-	EXIF_LOADER_READ,
-	EXIF_LOADER_READ_SIZE_HIGH_BYTE,
-	EXIF_LOADER_READ_SIZE_LOW_BYTE,
-	EXIF_LOADER_SKIP_BYTES,
-	EXIF_LOADER_EXIF_FOUND,
-	EXIF_LOADER_FAILED
-} ExifLoaderState;
-
-typedef struct {
-	ExifLoaderState state;
-	ExifData *data;
-	int      size;
-	int      last_marker;
-	guchar  *buffer_data;
-	int      bytes_read;
-} ExifLoaderData;
-
-/* This function imitates code from libexif, written by Lutz
- * Mller. See libexif/exif-data.c:exif_data_new_from_file. Here, it
- * can cope with a sequence of data chunks.
- */
-static gboolean
-exif_loader_write (ExifLoaderData *eld, guchar *buffer, int len)
-{
-	int i;
-	int len_remain;
-
-	if (eld->state == EXIF_LOADER_FAILED) return FALSE;
-	if (eld->state == EXIF_LOADER_EXIF_FOUND && eld->data != NULL) return FALSE;
-
-	for (i = 0; (i < len) && eld->state != EXIF_LOADER_EXIF_FOUND && eld->state != EXIF_LOADER_FAILED; i++) {
-
-		switch (eld->state) {
-		case EXIF_LOADER_SKIP_BYTES:
-			eld->size--;
-			if (eld->size == 0) {
-				eld->state = EXIF_LOADER_READ;
-			}
-			break;
-			
-		case EXIF_LOADER_READ_SIZE_HIGH_BYTE:
-			eld->size = buffer [i] << 8;
-			eld->state = EXIF_LOADER_READ_SIZE_LOW_BYTE;
-			break;
-			
-		case EXIF_LOADER_READ_SIZE_LOW_BYTE:
-			eld->size |= buffer [i];
-			eld->size = eld->size - 2;
-			
-			switch (eld->last_marker) {
-			case JPEG_MARKER_APP0:
-				eld->state = EXIF_LOADER_SKIP_BYTES;
-				break;
-
-			case JPEG_MARKER_APP1:
-				eld->state = EXIF_LOADER_EXIF_FOUND;
-				break;
-
-			default:
-				g_assert_not_reached ();
-			}
-
-			eld->last_marker = 0;
-			break;
-
-		default:
-			if (buffer[i] != 0xff) {
-				if (buffer [i] == JPEG_MARKER_APP0 ||
-				    buffer [i] == JPEG_MARKER_APP1) 
-				{
-					eld->state = EXIF_LOADER_READ_SIZE_HIGH_BYTE;
-					eld->last_marker = buffer [i];
-				}
-				else if (buffer [i] != JPEG_MARKER_SOI) {
-					eld->state = EXIF_LOADER_FAILED;
-				}
-			}
-		}
-
-	}
-	
-	len_remain = len - i;
-
-	if (eld->state == EXIF_LOADER_EXIF_FOUND && len_remain > 0) {
-
-		if (eld->buffer_data == NULL) {
-			eld->buffer_data = g_new0 (guchar, eld->size);
-			eld->bytes_read = 0;
-		}
-
-		if (eld->bytes_read < eld->size) {
-			int cp_len;
-			int rest;
-
-			/* the number of bytes we need to copy */
-			rest = eld->size - eld->bytes_read;
-			cp_len = MIN (rest, len_remain);
-			
-			g_assert ((cp_len + eld->bytes_read) <= eld->size);
-
-			/* copy memory */
-			memcpy (eld->buffer_data + eld->bytes_read, &buffer[i], cp_len);
-
-			eld->bytes_read += cp_len;
-		}
-
-		if (eld->bytes_read == eld->size) {
-			eld->data = exif_data_new_from_data (eld->buffer_data, eld->size);
-		}
-	}
-
-	return (eld->data != NULL);
-}
 
 #endif 
 
@@ -853,7 +736,7 @@ real_image_load (gpointer data)
 	gboolean failed;
 	GError *error = NULL;
 #if HAVE_EXIF
-	ExifLoaderData *eld;
+	ExifLoader *exif_loader;
 #endif
 
 	img = EOG_IMAGE (data);
@@ -870,6 +753,10 @@ real_image_load (gpointer data)
 		g_warning ("exif data not freed\n");
 	}
 #endif
+	if (priv->file_type != NULL) {
+		g_free (priv->file_type);
+		priv->file_type = NULL;
+	}
 
 	result = gnome_vfs_open_uri (&handle, priv->uri, GNOME_VFS_OPEN_READ);
 	if (result != GNOME_VFS_OK) {
@@ -899,13 +786,13 @@ real_image_load (gpointer data)
 	}
 	bytes_read_total = 0;
 	priv->progress = 0.0;
+	priv->bytes = info->size;
 	
 	buffer = g_new0 (guchar, READ_BUFFER_SIZE);
 	loader = gdk_pixbuf_loader_new ();
 	failed = FALSE;
 #if HAVE_EXIF
-	eld = g_new0 (ExifLoaderData, 1);
-	eld->state = EXIF_LOADER_READ;
+	exif_loader = exif_loader_new ();
 #endif
 
 	g_signal_connect_object (G_OBJECT (loader), "size-prepared", (GCallback) load_size_prepared, img, 0);
@@ -938,12 +825,14 @@ real_image_load (gpointer data)
 		}
 
 #if HAVE_EXIF
-		if (exif_loader_write (eld, buffer, bytes_read)) {
+		if (exif_loader != NULL && (exif_loader_write (exif_loader, buffer, bytes_read) == 0)) {
 			g_mutex_lock (img->priv->status_mutex);
-			exif_data_ref (eld->data);
-			priv->exif = eld->data;
+			priv->exif = exif_loader_get_data (exif_loader);
 			priv->load_status |= EOG_IMAGE_LOAD_STATUS_INFO_DONE;
 			g_mutex_unlock (img->priv->status_mutex);
+			
+			exif_loader_unref (exif_loader);
+			exif_loader = NULL;
 		}
 #endif
 	}
@@ -954,7 +843,7 @@ real_image_load (gpointer data)
 	gnome_vfs_close (handle);
 	gnome_vfs_file_info_unref (info);
 	
-	if (failed) {
+	if (failed || (bytes_read_total == 0)) {
 		g_mutex_lock (priv->status_mutex);
 		if (priv->image != NULL) {
 			g_object_unref (priv->image);
@@ -966,9 +855,13 @@ real_image_load (gpointer data)
 		if (error != NULL) {
 			priv->error_message = g_strdup (error->message);
 		}
+		else if (bytes_read_total == 0) {
+			priv->error_message = g_strdup (_("empty file"));
+		}
 		else {
 			priv->error_message = NULL;
 		}
+		priv->status = EOG_IMAGE_STATUS_FAILED;
 		g_mutex_unlock (priv->status_mutex);
 	}
 	else if (priv->cancel_loading) {
@@ -977,16 +870,24 @@ real_image_load (gpointer data)
 			g_object_unref (priv->image);
 			priv->image = NULL;
 		}
-		priv->cancel_loading = TRUE;
+#if HAVE_EXIF
+		if (priv->exif != NULL) {
+			exif_data_unref (priv->exif);
+			priv->exif = NULL;
+		}
+#endif
+		priv->cancel_loading = FALSE;
 		priv->progress = 0.0;
 
 		priv->load_status |= EOG_IMAGE_LOAD_STATUS_CANCELLED;
+		priv->status = EOG_IMAGE_STATUS_UNKNOWN;
 		g_mutex_unlock (priv->status_mutex);
 	}
 	else {
 		GdkPixbuf *image;
 		GdkPixbuf *transformed = NULL;
 		EogTransform *trans;
+		GdkPixbufFormat *format;
 
 		g_mutex_lock (priv->status_mutex);
 		image = priv->image;
@@ -1011,24 +912,24 @@ real_image_load (gpointer data)
 		priv->image = image;
 		priv->width = gdk_pixbuf_get_width (priv->image);
 		priv->height = gdk_pixbuf_get_height (priv->image);
+		format = gdk_pixbuf_loader_get_format (loader);
+		if (format != NULL) {
+			priv->file_type = g_strdup (gdk_pixbuf_format_get_name (format));
+		}
 #if HAVE_EXIF
 		update_exif_data (img);
 #endif 
+		eog_image_cache_add (img);
+
 		if (trans != NULL && priv->mode == EOG_IMAGE_LOAD_PROGRESSIVE) {
 			priv->load_status |= EOG_IMAGE_LOAD_STATUS_TRANSFORMED;
 		}
+
 		priv->load_status |= EOG_IMAGE_LOAD_STATUS_DONE;
+		priv->status = EOG_IMAGE_STATUS_LOADED;
 		g_mutex_unlock (priv->status_mutex);
 	}
 
-#if HAVE_EXIF
-	if (eld->buffer_data != NULL)
-		g_free (eld->buffer_data);
-	if (eld->data)
-		exif_data_unref (eld->data);
-	g_free (eld);
-#endif
-	
 	if (error != NULL) {
 		g_error_free (error);
 	}
@@ -1045,27 +946,21 @@ void
 eog_image_load (EogImage *img, EogImageLoadMode mode)
 {
 	EogImagePrivate *priv;
-	gboolean image_loaded;
-	gboolean thread_running;
 
 	g_return_if_fail (EOG_IS_IMAGE (img));
 
 	priv = EOG_IMAGE (img)->priv;
 
-	g_return_if_fail (priv->uri != NULL);
-
-	g_mutex_lock (priv->status_mutex);
-	image_loaded = priv->image != NULL;
-	thread_running = priv->load_thread != NULL;
-	g_mutex_unlock (priv->status_mutex);
-
-	if (image_loaded && !thread_running) {
+	if (priv->status == EOG_IMAGE_STATUS_LOADED) {
 		g_assert (priv->image != NULL);
+		eog_image_cache_reload (img);
 		g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_LOADING_FINISHED], 0);
 		g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_LOADING_INFO_FINISHED], 0);
 	}
-	else if (!image_loaded && !thread_running) { 
-                /* we assume that in all other cases there is no image loaded */
+	else if (priv->status == EOG_IMAGE_STATUS_FAILED) {
+		g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_LOADING_FAILED], 0, "");		
+	}
+	else if (priv->status == EOG_IMAGE_STATUS_UNKNOWN) { 
 		g_assert (priv->image == NULL);
 		
                 /* make sure the object isn't destroyed while we try to load it */
@@ -1115,9 +1010,9 @@ eog_image_load (EogImage *img, EogImageLoadMode mode)
 		}
 
 		/* start the thread machinery */
-		priv->load_id = g_timeout_add (CHECK_LOAD_TIMEOUT, check_load_status, img);
+		priv->status      = EOG_IMAGE_STATUS_LOADING;
+		priv->load_id     = g_timeout_add (CHECK_LOAD_TIMEOUT, check_load_status, img);
 		priv->load_thread = g_thread_create (real_image_load, img, TRUE, NULL);
-		
 	}
 }
 
@@ -1318,15 +1213,78 @@ eog_image_undo (EogImage *img)
 	priv->modified = (priv->undo_stack != NULL);
 }
 
+/* is_local_uri: 
+ * 
+ * Checks if the URI points to a local file system. This tests simply
+ * if the URI scheme is 'file'. This function is used to ensure that
+ * we can write to the path-part of the URI with non-VFS aware
+ * filesystem calls.
+ */
+static gboolean 
+is_local_uri (const GnomeVFSURI* uri)
+{
+	const char *scheme;
+
+	g_return_val_if_fail (uri != NULL, FALSE);
+
+	scheme = gnome_vfs_uri_get_scheme (uri);
+	return (g_strcasecmp (scheme, "file") == 0);
+}
+
+static char*
+get_save_file_type_by_suffix (const char *local_path)
+{
+	GSList *savable_formats = NULL;
+	GSList *it;
+	char *file_type = NULL;
+	char *suffix;
+	GdkPixbufFormat *format;
+	char **extension;
+	int i;
+
+	/* FIXME: this is probably not unicode friendly */
+	suffix = g_strrstr (local_path, ".");
+	if (suffix == NULL) {
+		return NULL;
+	}
+
+	/* skip '.' from the suffix string */
+	if (strlen (suffix) > 1) {
+		suffix++;
+	}
+
+	savable_formats = eog_pixbuf_get_savable_formats ();
+
+	/* iterate over the availabe formats and check for every
+	 * possible format suffix if it matches the file suffix */
+	for (it = savable_formats; it != NULL && file_type == NULL; it = it->next) {
+
+		format = (GdkPixbufFormat*) it->data;
+		extension = gdk_pixbuf_format_get_extensions (format);
+		
+		for (i = 0; extension[i] != NULL && file_type == NULL; i++) {
+			if (g_ascii_strcasecmp (extension[i], suffix) == 0) {
+				file_type = gdk_pixbuf_format_get_name (format);
+			}
+		}
+
+		g_strfreev (extension);
+	}
+	g_slist_free (savable_formats);
+
+	/* file_type is either NULL or contains the name of the format */
+	return file_type;
+}
+
 gboolean
-eog_image_save (EogImage *img, const GnomeVFSURI *uri, GError **error)
+eog_image_save (EogImage *img, GnomeVFSURI *uri, GdkPixbufFormat *format, GError **error)
 {
 	EogImagePrivate *priv;
 	char *file;
-	char *file_type = NULL;
-	GSList *savable_formats = NULL;
-	GSList *it;
-	gboolean result;
+	char *target_type = NULL;
+	char *source_type = NULL;
+	gboolean result = FALSE;
+	gboolean source_is_local;
 
 	g_return_val_if_fail (EOG_IS_IMAGE (img), FALSE);
 	g_return_val_if_fail (uri != NULL, FALSE);
@@ -1334,6 +1292,7 @@ eog_image_save (EogImage *img, const GnomeVFSURI *uri, GError **error)
 
 	priv = img->priv;
 
+	/* fail if there is no image to save */
 	if (priv->image == NULL) {
 		g_set_error (error, EOG_IMAGE_ERROR,
 			     EOG_IMAGE_ERROR_NOT_LOADED,
@@ -1341,85 +1300,181 @@ eog_image_save (EogImage *img, const GnomeVFSURI *uri, GError **error)
 		return FALSE;
 	}
 
-#if 0
-	/* FIXME: Why does this not work for local files? */
-	if (!gnome_vfs_uri_is_local (uri)) {
+	/* Since we rely on some local system calls resp. non-VFS aware libs, 
+	 * we can write only to local files.
+	 */
+	if (!is_local_uri (uri)) {
 		g_set_error (error, EOG_IMAGE_ERROR,
 			     EOG_IMAGE_ERROR_NOT_LOADED,
 			     _("Images can only be saved as local files."));
 		return FALSE;
 	}
-#endif
-	
-	/* find file type for saving, according to filename suffix */
-	file = (char*) gnome_vfs_uri_get_path (uri); /* don't free file */
-	
-	savable_formats = eog_pixbuf_get_savable_formats ();
-	for (it = savable_formats; it != NULL && file_type == NULL; it = it->next) {
-		GdkPixbufFormat *format;
-		char **extension;
-		int i;
 
-		format = (GdkPixbufFormat*) it->data;
-		extension = gdk_pixbuf_format_get_extensions (format);
-		
-		for (i = 0; extension[i] != NULL && file_type == NULL; i++) {
-			char *suffix;
-			
-			suffix = g_strconcat (".", extension[i], NULL);
-			if (g_str_has_suffix (file, suffix)) {
-				file_type = gdk_pixbuf_format_get_name (format);
-			}
+	file = (char*) gnome_vfs_uri_get_path (uri); /* don't free file, it's a static string */
 
-			g_free (suffix);
+	/* determine type of file to write (eg. 'png') */
+	if (format == NULL) {
+		if (gnome_vfs_uri_equal (priv->uri, uri)) {
+			/* we overwrite the same file, therefore we
+			 * can reuse the file type saved in the EogImage object. */
+			target_type = g_strdup (priv->file_type);
+		}
+		else {
+			/* find file type for saving, according to filename suffix */
+			target_type = get_save_file_type_by_suffix (file);
 		}
-
-		g_strfreev (extension);
 	}
-	g_slist_free (savable_formats);
-	
-	if (file_type == NULL) {
+	else {
+		target_type = gdk_pixbuf_format_get_name (format);
+	}
+
+	if (target_type == NULL) {
 		g_set_error (error, GDK_PIXBUF_ERROR,
 			     GDK_PIXBUF_ERROR_UNKNOWN_TYPE,
 			     _("Unsupported image type for saving."));
 		return FALSE;
 	}
+	
+	source_is_local = (priv->uri != NULL && is_local_uri (priv->uri));
+	source_type = priv->file_type;
+	
+	/* Check for some special cases, so that we use always the
+	 * least intrusive method for saving the image.
+	 */
+	if ((g_ascii_strcasecmp (target_type, source_type) == 0) &&
+	    !eog_image_is_modified (img))
+	{
+		/* If image source and target have the same type and
+		 * source is not modified in any way => copy file.
+		 */
+		GnomeVFSResult vfs_result;
+		
+		vfs_result = gnome_vfs_xfer_uri (priv->uri,
+						 uri, 
+						 GNOME_VFS_XFER_DEFAULT                 /* copy the data */,
+						 GNOME_VFS_XFER_ERROR_MODE_ABORT,       /* abort on all errors */
+						 GNOME_VFS_XFER_OVERWRITE_MODE_REPLACE, /* we checked for existing 
+											   file already */
+						 NULL,                                  /* no progress callback */
+						 NULL);
+		result = (vfs_result == GNOME_VFS_OK);
+		if (!result) {
+			g_set_error (error, EOG_IMAGE_ERROR,
+				     EOG_IMAGE_ERROR_VFS, 
+				     gnome_vfs_result_to_string (vfs_result));
+		}
+		g_print ("copy image file\n");
+	}
+#if HAVE_JPEG
+	else if ((g_ascii_strcasecmp (source_type, "jpeg") == 0) &&
+		 (g_ascii_strcasecmp (target_type, "jpeg") == 0) && 
+		 source_is_local) 
+	{
+		/* If source is a local file and both are jpeg files,
+		 * then use lossless transformation through libjpeg.
+		 */
+		result = eog_image_jpeg_save_lossless (img, uri, error);
 
+		g_print ("loseless saving of %s to %s\n", 
+			 gnome_vfs_uri_to_string (priv->uri, GNOME_VFS_URI_HIDE_NONE), file);
+	}
 #if HAVE_EXIF
-	if ((g_ascii_strcasecmp (file_type, "jpeg") == 0) && priv->exif != NULL) {
-		result = eog_image_jpeg_save (img, file, error);
+	else if ((g_ascii_strcasecmp (target_type, "jpeg") == 0) && priv->exif != NULL) {
+		/* If the target is a jpeg file and the image has EXIF
+		 * information preserve these by saving through
+		 * libjpeg.
+		 */
+		g_print ("Save through jpeg library.\n");
+		result = eog_image_jpeg_save (img, uri, error);
 	}
-	else {
-		result = gdk_pixbuf_save (priv->image, file, file_type, error, NULL);
-	}
-#else
-	result = gdk_pixbuf_save (priv->image, file, file_type, error, NULL);
 #endif
-
-
+#endif
+	else {
+		/* In all other cases: Use default save method
+		 * provided by gdk-pixbuf library.
+		 */
+		g_print ("default save method.\n");
+		result = gdk_pixbuf_save (priv->image, file, target_type, error, NULL);
+	}
+	
 	if (result) {
 		/* free the transformation since it's not needed anymore */
 		/* FIXME: You can't undo prev transformations then anymore. */
 		GList *it = priv->undo_stack;
 		for (; it != NULL; it = it->next) 
 			g_object_unref (G_OBJECT (it->data));
-
+		
 		g_list_free (priv->undo_stack);
-		g_object_unref (priv->trans);
-		priv->trans = NULL;
 		priv->undo_stack = NULL;
+		if (priv->trans != NULL) {
+			g_object_unref (priv->trans);
+			priv->trans = NULL;
+		}
 		priv->modified = FALSE;
+
+		/* update file properties */
+		if (priv->uri != NULL) {
+			gnome_vfs_uri_unref (priv->uri);
+		}
+		priv->uri = gnome_vfs_uri_ref (uri);
+		if (priv->caption != NULL) {
+			g_free (priv->caption);
+			priv->caption = NULL;
+		}
+		if (priv->caption_key != NULL) {
+			g_free (priv->caption_key);
+			priv->caption_key = NULL;
+		}
+		if (priv->file_type != NULL) {
+			g_free (priv->file_type);
+		}
+		priv->file_type = g_strdup (target_type);
+
+		g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_IMAGE_CHANGED], 0);
 	}
 
-	g_free (file_type);
+	g_free (target_type);
 
 	return result;
 }
 
+/*
+ * This function is extracted from 
+ * File: nautilus/libnautilus-private/nautilus-file.c
+ * Revision: 1.309
+ * Author: Darin Adler <darin@bentspoon.com>
+ */
+static gboolean
+have_broken_filenames (void)
+{
+	static gboolean initialized = FALSE;
+	static gboolean broken;
+	
+	if (initialized) {
+		return broken;
+	}
+	
+	broken = g_getenv ("G_BROKEN_FILENAMES") != NULL;
+  
+	initialized = TRUE;
+  
+	return broken;
+}
+
+/* 
+ * This function is inspired by
+ * nautilus/libnautilus-private/nautilus-file.c:nautilus_file_get_display_name_nocopy
+ * Revision: 1.309
+ * Author: Darin Adler <darin@bentspoon.com>
+ */
 gchar*               
 eog_image_get_caption (EogImage *img)
 {
 	EogImagePrivate *priv;
+	char *name;
+	char *utf8_name;
+	gboolean validated = FALSE;
+	gboolean broken_filenames;
 
 	g_return_val_if_fail (EOG_IS_IMAGE (img), NULL);
 
@@ -1427,6 +1482,48 @@ eog_image_get_caption (EogImage *img)
 
 	if (priv->uri == NULL) return NULL;
 
+	if (priv->caption != NULL) 
+		/* Use cached caption string */
+		return priv->caption;
+
+	name = gnome_vfs_uri_extract_short_name (priv->uri);
+	
+	if (name != NULL && gnome_vfs_uri_is_local (priv->uri)) {
+		/* Support the G_BROKEN_FILENAMES feature of
+		 * glib by using g_filename_to_utf8 to convert
+		 * local filenames to UTF-8. Also do the same
+		 * thing with any local filename that does not
+		 * validate as good UTF-8.
+		 */
+		broken_filenames = have_broken_filenames ();
+		if (broken_filenames || !g_utf8_validate (name, -1, NULL)) {
+			utf8_name = g_locale_to_utf8 (name, -1, NULL, NULL, NULL);
+			if (utf8_name != NULL) {
+				g_free (name);
+				name = utf8_name;
+				/* Guaranteed to be correct utf8 here */
+				validated = TRUE;
+			}
+		} 
+		else if (!broken_filenames) {
+			/* name was valid, no need to re-validate */
+			validated = TRUE;
+		}
+	}
+	
+	if (!validated && !g_utf8_validate (name, -1, NULL)) {
+		if (name == NULL) {
+			name = "[Invalid Unicode]";
+		}
+		else {
+			utf8_name = eel_make_valid_utf8 (name);
+			g_free (name);
+			name = utf8_name;
+		}
+	}
+
+	priv->caption = name;
+
 	if (priv->caption == NULL) {
 		char *short_str;
 
@@ -1443,30 +1540,46 @@ eog_image_get_caption (EogImage *img)
 	return priv->caption;
 }
 
-void
-eog_image_free_mem (EogImage *img)
+void 
+eog_image_free_mem_private (EogImage *image)
 {
 	EogImagePrivate *priv;
 	
-	g_return_if_fail (EOG_IS_IMAGE (img));
-	
-	priv = img->priv;
+	priv = image->priv;
 
-	if (priv->image != NULL) {
-		gdk_pixbuf_unref (priv->image);
-		priv->image = NULL;
+	if (priv->status == EOG_IMAGE_STATUS_LOADING) {
+		eog_image_cancel_load (image);
 	}
-
+	else {
+		if (priv->image != NULL) {
+			gdk_pixbuf_unref (priv->image);
+			priv->image = NULL;
+		}
+		
 #if HAVE_EXIF
-	if (priv->exif != NULL) {
-		exif_data_unref (priv->exif);
-		priv->exif = NULL;
-	}
+		if (priv->exif != NULL) {
+			exif_data_unref (priv->exif);
+			priv->exif = NULL;
+		}
 #endif
 
-	priv->load_status = EOG_IMAGE_LOAD_STATUS_NONE;
+		priv->load_status = EOG_IMAGE_LOAD_STATUS_NONE;
+		priv->status = EOG_IMAGE_STATUS_UNKNOWN;
+	}
 }
 
+void
+eog_image_free_mem (EogImage *image)
+{
+	g_return_if_fail (EOG_IS_IMAGE (image));
+
+	if (image->priv->image != NULL) {
+		eog_image_cache_remove (image);
+		eog_image_free_mem_private (image);
+	}
+}
+
+
 const gchar*        
 eog_image_get_collate_key (EogImage *img)
 {
@@ -1490,21 +1603,16 @@ void
 eog_image_cancel_load (EogImage *img)
 {
 	EogImagePrivate *priv;
-	gboolean join_thread = FALSE;
 
 	g_return_if_fail (EOG_IS_IMAGE (img));
 	
 	priv = img->priv;
 
 	g_mutex_lock (priv->status_mutex);
-	if (priv->load_thread != NULL) {
+	if (priv->status == EOG_IMAGE_STATUS_LOADING) {
 		priv->cancel_loading = TRUE;
-		join_thread = TRUE;
 	}
 	g_mutex_unlock (priv->status_mutex);
-
-	if (join_thread)
-		g_thread_join (priv->load_thread);
 }
 
 gpointer
@@ -1538,7 +1646,7 @@ eog_image_is_loaded (EogImage *img)
 	priv = img->priv;
 
 	g_mutex_lock (priv->status_mutex);
-	result = (priv->load_thread == NULL && priv->image != NULL);
+	result = (priv->status == EOG_IMAGE_STATUS_LOADED);
 	g_mutex_unlock (priv->status_mutex);
 	
 	return result;
@@ -1559,3 +1667,11 @@ eog_image_is_modified (EogImage *img)
 	
 	return img->priv->modified;
 }
+
+GnomeVFSFileSize
+eog_image_get_bytes (EogImage *img)
+{
+	g_return_val_if_fail (EOG_IS_IMAGE (img), 0);
+	
+	return img->priv->bytes;
+}
